# -*- coding: utf-8 -*-
"""
Django Tagulous - Fabulous Tags
"""
from __future__ import unicode_literals


__version__ = "0.14.1"
__license__ = "BSD"
__author__ = "Richard Terry"
__credits__ = [
    "Pamela McA'Nulty (PamelaM)",
    "James Pic (jpic)",
    "Robert Erb (rerb) at AASHE",
    "Mary (minidietcoke)",
    "Gaël Utard (gutard)",
    "Mark London",
    "Martín R. Guerrero (slackmart)",
    "Peter Baumgartner (ipmb)",
    "Raniere Silva (rgaics)",
    "Dmitry Ivanchenko (ivanchenkodmitry)",
]
